
class OSVReproducerError(Exception):
    """Generic errors."""
    pass


class OSVError(Exception):
    """Generic errors."""
    pass


class GitHubError(Exception):
    """Generic errors."""
    pass


class BuildError(Exception):
    """Generic errors."""
    pass


class RunnerError(Exception):
    """Generic errors."""
    pass


class OSSFuzzError(Exception):
    """Generic errors."""
    pass


class GCSError(Exception):
    """Generic errors."""
    pass


class DockerError(Exception):
    """Generic errors."""
    pass